/*

xsynth_tilde.c - NMF based cross-synthesis

*/

#include "m_pd.h"
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <lapacke/lapacke.h>

#define MAXPERMUTE 100

typedef struct _xsynth_tilde
{
    t_object x_obj;
    t_float x_f;
    t_outlet *x_out1;
    t_outlet *x_out2;

    int x_debug;            /* debug flag */
    int x_cancomplain;
    t_symbol *x_inname;     /* table holding input */
    t_symbol *x_hname;      /* ... and "H" output */
    t_symbol *x_wname;      /* ... and "W" output */
    int x_nrows;
    int x_ncolumns;
    int x_nbasis;
    int x_permute[MAXPERMUTE];
    double x_permuteval[MAXPERMUTE];
} t_xsynth_tilde;

static t_class *xsynth_tilde_class;

    /* multiply matrix a by b, result into out.
    a is outrow x sharedim, b is sharedim x outcol,
        and out is outrow by outcol */
static void mpy_a_b(double *out, double *a, double *b, 
    int outrow, int outcol, int sharedim)
{
    int i, j, k;
    for (i = 0; i < outrow; i++)
        for (j = 0; j < outcol; j++)
    {
        double accum = 0;
        for (k = 0; k < sharedim; k++)
            accum += a[i * sharedim + k] * b[k * outcol + j];
        out[i * outcol + j] = accum;
    }
}

    /* multiply matrix (a transpose) by b, result into out.
    a is sharedim x outrow, b is sharedim x outcol,
        and out is outrow by outcol */
static void mpy_at_b(double *out, double *a, double *b, 
    int outrow, int outcol, int sharedim)
{
    int i, j, k;
    for (i = 0; i < outrow; i++)
        for (j = 0; j < outcol; j++)
    {
        double accum = 0;
        for (k = 0; k < sharedim; k++)
            accum += a[k * outrow + i] * b[k * outcol + j];
        out[i * outcol + j] = accum;
    }
}

    /* multiply matrix a by (b transpose), result into out.
    a is outrow x sharedim, b is outcol x sharedim,
        and out is outrow by outcol */
static void mpy_a_bt(double *out, double *a, double *b, 
    int outrow, int outcol, int sharedim)
{
    int i, j, k;
    for (i = 0; i < outrow; i++)
        for (j = 0; j < outcol; j++)
    {
        double accum = 0;
        for (k = 0; k < sharedim; k++)
            accum += a[i * sharedim + k] * b[j * sharedim + k];
        out[i * outcol + j] = accum;
    }
}

static void xsynth_tilde_debug(t_xsynth_tilde *x, t_float f)
{
    x->x_debug = f;
}

static void xsynth_tilde_print(t_xsynth_tilde *x, t_float f)
{
    int i, j;
    post("debug %d", x->x_debug);
}

static void xsynth_tilde_permute(t_xsynth_tilde *x, t_symbol *s,
    int argc, t_atom *argv)
{
    int i, j;
    if (argc > MAXPERMUTE)
        argc = MAXPERMUTE;
    for (i = 0; i < MAXPERMUTE; i++)
        x->x_permute[i] = i;
    for (i = 0; i < argc; i++)
        x->x_permute[i] = atom_getintarg(i, argc, argv);
}

static t_int *xsynth_tilde_perform(t_int *w)
{
    int i, j, laout;
    t_sample *in = (t_sample *)(w[1]);
    t_sample *out= (t_sample *)(w[2]);
    int n = (int)(w[3]);
    t_xsynth_tilde *x = (t_xsynth_tilde *)(w[4]);
    t_garray *ingarray, *hgarray, *wgarray;
    int insize, wsize, hsize;
    t_word *invec, *wvec, *hvec;
    int nrows = x->x_nrows, ncolumns = x->x_ncolumns, nbasis = x->x_nbasis;
    double *hcopy;      /* nbasis*ncolumns */
    double *hhttmp;     /* H * H transpose: nbasis*nbasis */
    double *testin;     /* test input, ncolumns*/
    double *testhttmp;   /* test * H transpose: 1*nbasis */
    double *outtmp;     /* reconstructed output: 1*ncolumns */
    lapack_int *lpiv;   /* stuff for PAPACK, nbasis ints */
    
    hcopy = (double *)alloca(
        sizeof(double) * (nbasis*ncolumns + nbasis*nbasis + 
            ncolumns + nbasis + ncolumns) +
        sizeof(lapack_int) * nbasis);
    hhttmp = hcopy + nbasis*ncolumns;
    testin = hhttmp + nbasis*nbasis;
    testhttmp = testin + ncolumns;
    outtmp = testhttmp + nbasis;
    lpiv = (lapack_int *)(outtmp + ncolumns);
    
    if (!(ingarray = (t_garray *)pd_findbyclass(x->x_inname,
            garray_class)) ||
                !garray_getfloatwords(ingarray, &insize, &invec))
    {
        if (x->x_cancomplain)
            post("%s: not found", x->x_inname->s_name);
        x->x_cancomplain = 0;
        goto foo;
    }
    if (!(wgarray = (t_garray *)pd_findbyclass(x->x_wname,
            garray_class)) ||
                !garray_getfloatwords(wgarray, &wsize, &wvec))
    {
        if (x->x_cancomplain)
            post("%s: not found", x->x_wname->s_name);
        x->x_cancomplain = 0;
        goto foo;
    }
    if (!(hgarray = (t_garray *)pd_findbyclass(x->x_hname,
            garray_class)) ||
                !garray_getfloatwords(hgarray, &hsize, &hvec))
    {
        if (x->x_cancomplain)
            post("%s: not found", x->x_hname->s_name);
        x->x_cancomplain = 0;
        goto foo;
    }
    if (insize < x->x_nrows * x->x_ncolumns)
    {
        if (x->x_cancomplain)
            post("table %s with %d elements too small to hold %dx%d matrix",
                x->x_inname->s_name, insize, x->x_nrows, x->x_ncolumns);
        x->x_cancomplain = 0;
        goto foo;
    }
    if (hsize < x->x_nbasis * x->x_ncolumns)
    {
        if (x->x_cancomplain)
            post("table %s with %d elements too small to hold %dx%d matrix",
                x->x_inname->s_name, hsize, x->x_nbasis, x->x_ncolumns);
        x->x_cancomplain = 0;
        goto foo;
    }
    if (wsize < x->x_nrows * x->x_nbasis)
    {
        if (x->x_cancomplain)
            post("table %s with %d elements too small to hold %dx%d matrix",
                x->x_inname->s_name, wsize, x->x_nrows, x->x_nbasis);
        x->x_cancomplain = 0;
        goto foo;
    }
    if (ncolumns != n)
    {
        if (x->x_cancomplain)
            post("ncolumns %d unequal to block size %d", ncolumns, n);
        x->x_cancomplain = 0;
        goto foo;
    }

    for (i = 0; i < nbasis*ncolumns; i++)
        hcopy[i] = hvec[i].w_float;
    for (i = 0; i < ncolumns; i++)
        testin[i] = in[i];
    mpy_a_bt(hhttmp, hcopy, hcopy, nbasis, nbasis, ncolumns);
    mpy_a_b(testhttmp, hcopy, testin, nbasis, 1, ncolumns);
    laout = LAPACKE_dsysv(LAPACK_COL_MAJOR, 'u',
        nbasis, 1, hhttmp, nbasis, lpiv, testhttmp, nbasis);
    if (laout != 0)
    {
        post("lapack returned error %d", laout);
        goto foo;
    }


    if (x->x_debug == 2)
    {
        int perm[100], ngood = 0;
        double newhttmp[10000], newtesthttmp[100];
        if (nbasis > 100)
        {
            post("basis <= 100 only");
            goto nope;
        }
        for (i = 0; i < nbasis; i++)
        {
            if (testhttmp[i] > 0)
                perm[ngood++] = i;
        }
        if (ngood <= 0)
            goto foo;
        for (i = 0; i < ngood; i++)
            for (j = 0; j < ngood; j++)
                newhttmp[i * ngood + j] = hhttmp[perm[i] * nbasis + perm[j]];
        mpy_a_b(newtesthttmp, hcopy, testin, nbasis, 1, ncolumns);
        for (i = 0; i < ngood; i++)
            newtesthttmp[i] = newtesthttmp[perm[i]];
        laout = LAPACKE_dsysv(LAPACK_COL_MAJOR, 'u',
            ngood, 1, (double *)newhttmp, ngood, lpiv, newtesthttmp, ngood);
        if (laout != 0)
        {
            post("lapack returned error %d", laout);
            goto foo;
        }
        for (i = 0; i < nbasis; i++)
            testhttmp[i] = 0;
        for (i = 0; i < ngood; i++)
            testhttmp[perm[i]] = newtesthttmp[i];
        
    }
nope:
    if (x->x_debug >= 1)
    {
        for (i = 0; i < nbasis; i++)
            if (testhttmp[i] < 0)
                testhttmp[i] = 0;
    }
    
        /* optionally permute the weights */
    j = (nbasis > MAXPERMUTE ? MAXPERMUTE : nbasis);
    for (i = 0; i < j; i++)
    {
        int chan = x->x_permute[i];
        if (chan >= 0 && chan < j)
            x->x_permuteval[i] = testhttmp[chan];
        else x->x_permuteval[i] = 0;
    }
    for (i = 0; i < j; i++)
        testhttmp[i] = x->x_permuteval[i];

    mpy_at_b(testin, hcopy, testhttmp, ncolumns, 1, nbasis);
    for (i = 0; i < ncolumns; i++)
        out[i] = testin[i];
    
foo:
    return (w+5);
}

static void xsynth_tilde_dsp(t_xsynth_tilde *x, t_signal **sp)
{
    dsp_add(xsynth_tilde_perform, 4,
        sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n, x);
    x->x_cancomplain = 1;
}

static void xsynth_tilde_free(t_xsynth_tilde *x)
{
}

static void *xsynth_tilde_new(t_symbol *s, int argc, t_atom *argv)
{
    t_xsynth_tilde *x = (t_xsynth_tilde *)pd_new(xsynth_tilde_class);
    int i;
    if (argc < 6 || argv[0].a_type != A_SYMBOL
        || argv[1].a_type != A_SYMBOL || argv[2].a_type != A_SYMBOL ||
            (x->x_nrows = atom_getintarg(3, argc, argv)) <= 0 ||
            (x->x_ncolumns = atom_getintarg(4, argc, argv)) <= 0 ||
            (x->x_nbasis = atom_getintarg(5, argc, argv)) <= 0)
    {
        post("usage: synth intab htab wtab rows columns nbasis");
        x->x_inname = x->x_wname = x->x_hname = &s_;
    }
    else
    {
        x->x_inname = atom_getsymbolarg(0, argc, argv);
        x->x_wname = atom_getsymbolarg(1, argc, argv);
        x->x_hname = atom_getsymbolarg(2, argc, argv);
    }
    x->x_f = 0;
    x->x_cancomplain = 1;
    outlet_new(&x->x_obj, gensym("signal"));
    x->x_out1 = outlet_new(&x->x_obj, &s_signal);
    x->x_debug = 0;
    for (i = 0; i < MAXPERMUTE; i++)
        x->x_permute[i] = i;
    return (x);
}

void xsynth_tilde_setup(void)
{
    xsynth_tilde_class = class_new(gensym("xsynth~"),
        (t_newmethod)xsynth_tilde_new, (t_method)xsynth_tilde_free,
	sizeof(t_xsynth_tilde), 0, A_GIMME, 0);
    CLASS_MAINSIGNALIN(xsynth_tilde_class, t_xsynth_tilde, x_f);
    class_addmethod(xsynth_tilde_class, (t_method)xsynth_tilde_dsp,
        gensym("dsp"), 0);

    class_addmethod(xsynth_tilde_class, (t_method)xsynth_tilde_permute,
        gensym("permute"), A_GIMME, 0);
    class_addmethod(xsynth_tilde_class, (t_method)xsynth_tilde_print,
        gensym("print"), A_DEFFLOAT, 0);
    class_addmethod(xsynth_tilde_class, (t_method)xsynth_tilde_debug,
        gensym("debug"), A_FLOAT, 0);

}

